use crate::model::*;

use std::collections::*;
use std::fs;
use std::io;
use std::path::{Path, PathBuf};
use std::sync::Arc;

pub use self::search::Query;
mod image_convert;
mod metadata_file;
mod search;

pub type TextureFileResult = Result<Arc<Vec<u8>>, TextureFileError>;
pub enum TextureFileError {
    NotFound,
    IoError(io::Error),
    ImageError(::image::ImageError),
}

impl From<io::Error> for TextureFileError {
    fn from(err: io::Error) -> Self {
        TextureFileError::IoError(err)
    }
}

impl From<::image::ImageError> for TextureFileError {
    fn from(err: ::image::ImageError) -> Self {
        TextureFileError::ImageError(err)
    }
}

pub struct DataStore {
    //data_path: PathBuf,
    base_dir: PathBuf,

    textures: HashSet<Texture>,

    id_index: HashMap<String, Texture>,
    name_index: HashMap<String, Texture>,

    preview_cache: HashMap<(TextureFormat, Sha256), Arc<Vec<u8>>>,
}

impl DataStore {
    fn texture_file_path(&self, sha: &Sha256) -> PathBuf {
        let mut path = self.base_dir.clone();
        path.push("textures");
        path.push(sha.as_hex_string());
        path
    }

    fn index_file_path(&self) -> PathBuf {
        let mut path = self.base_dir.clone();
        path.push("collection.json");
        path
    }

    pub fn new(base_dir: &Path) -> io::Result<DataStore> {
        let mut store = DataStore {
            base_dir: base_dir.into(),

            textures: Default::default(),
            id_index: Default::default(),
            name_index: Default::default(),

            preview_cache: Default::default(),
        };

        let metadata_file = metadata_file::MetadataFile::load(base_dir)?;

        for texture in metadata_file.textures.iter() {
            match store.insert(texture.clone())? {
                true => (),
                false => {
                    println!("inserting {:#?} failed !!!", texture); // TODO: What should be done?
                }
            }
        }

        Ok(store)
    }

    pub fn texture_by_id(&self, id: &str) -> Option<Texture> {
        self.id_index.get(id).cloned()
    }

    pub fn texture_by_name(&self, id: &str) -> Option<Texture> {
        self.name_index.get(id).cloned()
    }

    pub fn insert(&mut self, texture: Texture) -> io::Result<bool> {
        if self.id_index.contains_key(&texture.id) {
            return Ok(false);
        }

        if self.name_index.contains_key(&texture.name) {
            return Ok(false);
        }

        if !self.is_texture_file_on_disk(&texture.texture_hash)? {
            return Ok(false);
        }

        self.id_index.insert(texture.id.clone(), texture.clone());
        self.name_index
            .insert(texture.name.clone(), texture.clone());
        self.textures.insert(texture.clone());

        Ok(true)
    }

    /// returns true if successful
    pub fn delete(&mut self, tex: &Texture) -> bool {
        if self.textures.remove(tex) {
            // remove
            assert!(self.id_index.remove(&tex.id).is_some());
            assert!(self.name_index.remove(&tex.name).is_some());

            // don't delete cache, since it could be used
            // by other texture.

            true
        } else {
            false
        }
    }

    /// Check if the texture, given by hash, physically exists on the file system.
    pub fn is_texture_file_on_disk(&self, hash: &Sha256) -> io::Result<bool> {
        let file_path = self.texture_file_path(&hash);

        Ok(file_path.is_file())
    }

    pub fn read_texture_file_by_hash(&mut self, hash: &Sha256) -> TextureFileResult {
        use std::fs::*;
        use std::io::*;

        if !(self.is_texture_file_on_disk(&hash)?) {
            return Err(TextureFileError::NotFound);
        }

        let file_path = self.texture_file_path(&hash);

        let mut file = File::open(file_path)?;
        let mut buffer = Vec::new();
        file.read_to_end(&mut buffer)?;

        Ok(Arc::new(buffer))
    }

    pub fn store_texture_file(&mut self, data: &[u8]) -> io::Result<()> {
        use std::fs;
        use std::io::Write;

        let hash = crate::model::Sha256::from_data(data);
        let file_path = self.texture_file_path(&hash);

        fs::create_dir_all(&file_path)?;

        let mut file = fs::File::create(&file_path)?;

        file.write_all(data)
    }

    fn flush_metadata(&self) -> io::Result<()> {
        use std::ops::Deref;

        let f = metadata_file::MetadataFile::from_iterator(self.textures.iter());

        f.store(self.index_file_path().as_path())
    }

    pub fn get_texture_preview(&mut self, hash: &Sha256) -> TextureFileResult {
        unimplemented!();
    }

    pub fn garbage_collect(&mut self) -> io::Result<()> {
        unimplemented!()
    }

    pub fn query(&mut self, query: &self::search::Query) -> Vec<Texture> {
        unimplemented!();
        // calls self::search::search(... )
    }
}
